import { point } from 'regef-geometry'
import Capability from './Capability'
import { START_CONNECTION, END_CONNECTION } from './constants'
import {
  eraseFeedback,
  requestFeedback,
  perform,
  partMatches,
  typeMatches,
  isLeftButton,
} from './utils'

const DEFAULT_CONFIG = {
  parts: null,
  sourceTypes: [],
  targetTypes: [],
}

export default class ConnectCapability extends Capability {
  constructor(engine, config = {}) {
    super(engine)
    this.config = { ...DEFAULT_CONFIG, ...config }
    this.init()
  }

  init() {
    this.progress = false
    this.source = null
    this.target = null
    this.coordinates = null
    this.lastRequest = null
  }

  cancel() {
    if (this.progress) {
      if (this.lastRequest !== null) {
        eraseFeedback(this.engine.editPolicies, this.lastRequest)
      }
      this.init()
    }
  }

  getStartConnectionRequest() {
    return {
      type: START_CONNECTION,
      source: this.source.component.userComponent,
      location: point(this.coordinates),
    }
  }

  getEndConnectionRequest() {
    return {
      type: END_CONNECTION,
      source: this.source.component.userComponent,
      target: this.target.component.userComponent,
      location: point(this.coordinates),
    }
  }

  buildEndConnectRequest(e) {
    const part = this.engine.domHelper.findPart(e.target, partMatches(this.config.parts))
    if (!part) {
      return null
    }
    const target = part.domHelper.findClosest(e.target, typeMatches(this.config.targetTypes))
    if (!target) {
      return null
    }
    this.target = target
    this.coordinates = point(e.clientX, e.clientY)
    return this.getEndConnectionRequest()
  }

  buildStartConnectionRequest(e) {
    const part = this.engine.domHelper.findPart(e.target, partMatches(this.config.parts))
    if (!part) {
      return null
    }
    const source = part.domHelper.findClosest(e.target, typeMatches(this.config.sourceTypes))
    if (!source) {
      return null
    }
    this.source = source
    this.coordinates = point(e.clientX, e.clientY)
    return this.getStartConnectionRequest()
  }

  handleFeedback(lastRequest, request) {
    if (lastRequest !== null && (request === null || request.target !== lastRequest.target)) {
      eraseFeedback(this.engine.editPolicies, lastRequest)
    }
    if (request !== null) {
      requestFeedback(this.engine.editPolicies, request)
    }
  }

  onMouseDown(e) {
    if (!isLeftButton(e)) {
      return
    }
    const request = this.buildStartConnectionRequest(e)
    if (request !== null) {
      this.progress = true
    }
    this.handleFeedback(this.lastRequest, request)
    this.lastRequest = request
  }

  onMouseMove(e) {
    if (!this.progress) {
      return
    }
    const request = this.buildEndConnectRequest(e)
    this.handleFeedback(this.lastRequest, request)
    this.lastRequest = request
  }

  onMouseUp(e) {
    if (!this.progress) {
      return
    }
    const request = this.buildEndConnectRequest(e)
    if (this.lastRequest !== null) {
      eraseFeedback(this.engine.editPolicies, this.lastRequest)
    }
    if (request !== null) {
      perform(this.engine.editPolicies, request)
    }
    this.init()
  }
}
